# encoding: utf-8

# mixin Graphviz name space (what ever it belongs)
module Graphviz
  # A name space for all types of diagram figures
  module Diagram
    # An UML class diagram
    class ClassDiagram
      attr_reader :graphviz

      # A entity class with many attributes and methods definition strings
      class Entity
        # The `attributes` and `methods` are lists of Hashes contain
        # information necessary to produce Graphviz labels, such as
        # `name`, `visibility`, `arguments` (for methods) and `type`.
        #
        # A key `field_id` will generated in a auto-increment manner
        # if a relationship (e.g., belongs to) are attached to a specific
        # member field.
        #
        # `node_attributes` are Graphviz attributes for the node represents
        # the entity.
        attr_accessor :name, :attributes, :methods, :node_attributes
        def initialize(name, attrs = {})
          @name, @attributes, @methods = name, [], []
          @node_attributes = { shape: 'record' }.merge(attrs)
        end

        def member(name)
          members.each { |m| return m if m[:name] == name }
        end

        def attribute(name)
          attributes.each { |m| return m if m[:name] == name }
        end

        def method(name)
          methods.each { |m| return m if m[:name] == name }
        end

        # All member fields include both attributes and methods
        def members
          @attributes + @methods
        end

        def add_attribute(name, opts = {})
          opts[:name] = name
          opts[:visibility] ||= :public
          opts.keys.each do |k|
            known_keys = [:name, :visibility, :type, :field_id]
            fail "unknown attribute #{k}" unless known_keys.include?(k.to_sym)
          end
          @attributes << opts
        end

        def add_method(name, opts = {})
          opts[:name] = name
          opts[:visibility] ||= :public
          opts.keys.each do |k|
            known_keys = [:name, :visibility, :arguments, :type, :field_id]
            fail "unknown attribute #{k}" unless known_keys.include?(k.to_sym)
          end
          @methods << opts
        end

        def belongs_to(entity, opts = {})
          Aggragation.new self, entity, opts
        end

        def embedded_in(entity, opts = {})
          Composition.new self, entity, opts
        end

        def extends(entity, opts = {})
          Generalization.new self, entity, opts
        end

        def implements(entity, opts = {})
          Realization.new self, entity, opts
        end

        # rubocop:disable MethodLength
        def label
          return name unless node_attributes[:shape].to_sym == :record
          rl = RecordLabel.new
          rl.add_row(name)
          rl.add_separator
          attributes.each do |at|
            str = format '%s%s', visibility_symbol(at), at[:name]
            str = "#{str} : #{at[:type].capitalize}" if at[:type]
            rl.add_row str, align: :left, field_id: at[:field_id]
          end

          rl.add_separator
          methods.each do |at|
            arguments = at[:arguments] ? at[:arguments] : ''
            str = format '%s%s(%s)', visibility_symbol(at), at[:name],
                         arguments.sub(/\A\(/, '').sub(/\)\Z/, '')
            str = "#{str} : #{at[:type].capitalize}" if at[:type]
            rl.add_row str, align: :left, field_id: at[:field_id]
          end

          rl.to_s
        end

        private

        def visibility_symbol(key)
          case key[:visibility].to_sym
          when :private then '- '
          when :protected then '# '
          when :derived then '/ '
          when :package then '~ '
          else
            '+ '
          end
        end
      end

      # Relationship between entities
      class Link
        # From and to are entities links each other
        attr_accessor :from, :to, :attributes

        def initialize(from, to, attributes = {})
          attrs = default_attrs.merge(attributes)
          @from, @to, @attributes = from, to, attrs
        end

        def default_attrs
          {}
        end
      end

      # :from belongs to :to
      class Aggragation < Link
        def default_attrs
          { arrowhead: 'diamond', headlabel: '1', taillabel: '0..*' }
        end
      end

      # :from embeds in :to
      class Composition < Link
        def default_attrs
          { arrowhead: 'odiamond', headlabel: '1', taillabel: '0..*' }
        end
      end

      # :from extends (sub-classing) :to
      class Generalization < Link
        def default_attrs
          { arrowhead: 'empty' }
        end
      end

      # :from implement :to (interface)
      class Realization < Link
        def default_attrs
          { style: 'dashed', arrowhead: 'vee', label: 'implements' }
        end
      end

      # All `opts` are passed to the constructor of `GraphViz`
      def initialize(opts = {})
        opts[:type] ||= :digraph
        opts[:rankdir] ||= 'LR'
        @graphviz = GraphViz.new :G, opts
        @entities, @links = [], []
      end

      def <<(obj)
        case obj
        when Link then @links << obj
        when Entity then @entities << obj
        else
          fail "unknown type #{obj.class}, can not add to graph"
        end
      end

      def add_entity(name, opts = {})
        self << Entity.new(name, opts)
      end

      def entity(name)
        @entities.each { |e| return e if e.name == name }
        nil
      end

      def [](name)
        add_entity(name) unless entity(name)
        entity(name)
      end

      def output(hsh)
        @entities.each do |e|
          attrs = { label: e.label }.merge(e.node_attributes)
          @graphviz.add_node(e.name, attrs)
        end

        @links.each do |l|
          @graphviz.add_edges(l.from.name, l.to.name, l.attributes)
        end

        @graphviz.output hsh
      end
    end
  end
end
